/*! Application-specific logic lives here

    **TODO:** Look into moving the argument definition into a
    [build.rs](https://doc.rust-lang.org/cargo/reference/build-scripts.html) like in the
    [clap_generate](https://docs.rs/clap_generate/3.0.0-beta.1/clap_generate/fn.generate_to.html)
    examples so I don't have build the completion generation code into the output binary.
*/

// Parts Copyright 2017-2021, Stephan Sokolow

// Standard library imports
// use std::path::PathBuf;
// use std::process;
use std::fs;
use std::collections::HashMap;

// 3rd-party crate imports
use anyhow::Result;
use structopt::StructOpt;

// Local Imports
use crate::helpers::{BoilerplateOpts, HELP_TEMPLATE};
// use crate::validators::path_readable_file;
use crate::clouds::{on_cloud};
use crate::defs::{BxDynResult};
use clds::clouds::utils::clear_specs;

use tkdr::crypt_lib::{encrypt, decrypt};
use tkdr::tera_lib::{hash_from_data, data_templated, hash_content};
use tkdr::randomkey::{RandomKey};
use key_of_life::get_key;

/// The verbosity level when no `-q` or `-v` arguments are given, with `0` being `-q`
pub const DEFAULT_VERBOSITY: u64 = 1;



/// Default KEY PATH 
pub const KEY_PATH: &str = ".k";
/// Command-line argument schema
///
/// ## Relevant Conventions:
///
///  * Make sure that there is a blank space between the `<name>` `<version>` line and the
///    description text or the `--help` output won't comply with the platform conventions that
///    `help2man` depends on to generate your manpage. (Specifically, it will mistake the `<name>
///    <version>` line for part of the description.)
///  * `StructOpt`'s default behaviour of including the author name in the `--help` output is an
///    oddity among Linux commands and, if you don't disable it, you run the risk of people
///    unfamiliar with `StructOpt` assuming that you are an egotistical person who made a conscious
///    choice to add it.
///
///    The proper standardized location for author information is the `AUTHOR` section which you
///    can read about by typing `man help2man`.
///
/// ## Cautions:
///  * Subcommands do not inherit `template` and it must be re-specified for each one.
///    ([clap-rs/clap#1184](https://github.com/clap-rs/clap/issues/1184))
///  * Double-check that your choice of `about` or `long_about` is actually overriding this
///    doc comment. The precedence has some bugs such as
///    [TeXitoi/structopt#391](https://github.com/TeXitoi/structopt/issues/391) and
///    [TeXitoi/structopt#333](https://github.com/TeXitoi/structopt/issues/333).
///  * Do not begin the description text for subcommands with `\n`. It will break the formatting in
///    the top-level help output's list of subcommands.
#[derive(StructOpt, Debug)]
#[structopt(template = HELP_TEMPLATE,
            about = "CloudMandala Declarative command line interface, to create and maintain open, free and sustainable Clouds",
            global_setting = structopt::clap::AppSettings::ColoredHelp)]
pub struct CliOpts {
    #[allow(clippy::missing_docs_in_private_items)] // StructOpt compile-time errors if we doc this
    #[structopt(flatten)]
    pub boilerplate: BoilerplateOpts,

    /// TskSrvc  to run 
    #[structopt(short, long)]
    pub tsksrvc: Option<String>,

    /// Command to run on TskSrvc
    #[structopt(short, long)]
    pub cmd: Option<String>,
  
    /// On next tsksrvc after TskSrvc selected ('next')
    #[structopt(short, long)]
    pub next: Option<String>,

    /// Source cloud path
    #[structopt(short, long)]
    pub source: Option<String>,

    ///  list of hosts
    #[structopt(short, long)]
    pub listhosts: Option<String>,

    /// Force running (-f, -ff etc.)
    #[structopt(short, long, parse(from_occurrences))]
    pub force: u8,
/*
    /// Encrypt text
    #[structopt(short, long)]
    pub encrypt: Option<String>,

    /// Encrypt file 
    #[structopt(long)]
    fencrypt: Option<String>,

    /// Decrypt text
    #[structopt(short, long)]
    decrypt: Option<String>,

    /// Decrypt file
    #[structopt(long)]
    fdecrypt: Option<String>,

    /// Random key (optional uuid format)
    #[structopt(short, long)]
    key: Option<String>,

    /// Get Blake3 hash from text
    #[structopt(long)]
    hash: Option<String>,

    /// Tera template with data
    #[structopt(short,long)]
    tera: Option<Vec<String>>,

    /// Get uuid v4 
    #[structopt(short,long)]
    uuid: Option<String>,

    /// Input for tsksrvc 
    #[structopt(short, long)]
    input: Option<String>,

    /// Output for tsksrvc 
    #[structopt(short, long)]
    output: Option<String>,

    /// Output for http 
    #[structopt(short, long)]
    http: Option<String>,
*/
    // File(s) to use as input
    //
    // **TODO:** Figure out if there's a way to only enforce constraints on this when not asking
    //           to dump completions.
    /*
    #[structopt(parse(from_os_str),
                validator_os = path_readable_file)  
                ]
    inpath: Vec<PathBuf>,
    */
    /// Target 
    #[structopt()]
args: Vec<String>,
}

/// The actual `main()`
/// 
/// Using a key and `XChaCha20Poly130` and random nonce
/// -run encrypt text-to-encrypt
/// -run decrypt text-to-decrypt
/// 
/// Format  -key n for a key with n characters long 
/// Format  -key n uuid  for n items in uuid format with dashes
#[allow(clippy::integer_arithmetic, clippy::restriction)]
pub async fn main(opts: CliOpts) -> BxDynResult<String> { // anyhow::Result<()> { 
  // println!("|{}|",&key);
  let key_path = &envmnt::get_or("KEY_PATH", KEY_PATH);
  let key = get_key(&key_path,None).await;
  // let key = get_key(KEY_PATH,None).await;
  if key.is_empty() {
    std::process::exit(0x0100);
  }
  let empty_res=String::from("");
  let mut run_tsksrvc = String::from("all");
  let mut run_cmd = String::from("");
  let mut run_next = String::from("");
  let mut listhosts = String::from("");
  if let Some(tsksrvc)= opts.tsksrvc {
    run_tsksrvc=String::from(&tsksrvc);
  }
  if let Some(cmd)=opts.cmd {
    run_cmd=String::from(&cmd);
  }
  if let Some(list_hosts)=opts.listhosts {
    listhosts=String::from(&list_hosts);
  }
	if let Some(nxt)=opts.next {
    run_next=String::from(&nxt);
  }
  if let Some(source)=opts.source {
    //let res = encrypt(&text, &key);
    match on_cloud(&run_tsksrvc, &run_cmd, &run_next, &source, &listhosts, opts.force).await {
      Ok(res) => return Ok(res),
      Err(e) => println!("Error on_cloud: {}",e),
    }
    if envmnt::get_or("KLDS_DEBUG", "").len() == 0 {
      println!("Clean on_cloud");
      match  clear_specs(&source).await {
        Ok(_) => return Ok(empty_res),
        Err(e) => println!("Error on_cloud: {}",e),
      }
    }
    let now = chrono::Utc::now().timestamp();
  //  println!("{}",&now);
    return Ok(empty_res);
  // TODO Chect to remove provision specs config 
//   println!("{} - {}",&tsksrvc,&source);
  }
  /*
  if opts.force > 0 {
    println!("Force: {}",opts.force);
    return Ok(());
  }
  // dbg!(&opts);
  if let Some(text)= opts.encrypt {
    let res = encrypt(&text, &key);
    println!("{}",&res);
    return Ok(());
  }
  if let Some(file)= opts.fencrypt {
    match fs::read_to_string(&file.as_str()) {
      Ok(content) => {
        let res = encrypt(&content, &key);
        println!("{}",&res);
      },
      Err(e) => {
        println!("Unable to open {}\n {}",&file,e);
      },
    }
   return Ok(());
  }
  if let Some(text)= opts.decrypt {
    let res = decrypt(&text, &key);
    println!("{}",&res);
    return Ok(());
  }
  if let Some(file)= opts.fdecrypt {
    match fs::read_to_string(&file.as_str()) {
      Ok(content) => {
        let res = decrypt(&content, &key);
        println!("{}",&res);
      },
      Err(e) => {
        println!("Unable to open {}\n {}",&file,e);
      },
    }
   return Ok(());
  }
  if core::option::Option::is_some(&opts.key) {
    let mut template = "";
    if let (Some(_),Some(templ)) = (opts.args.get(0),opts.args.get(1)) { 
       template = templ.as_str();
    }
    if let Some(text_long) = opts.args.get(0) {
      match text_long.parse::<u32>() { 
        Ok(lon) =>  {
          let random = RandomKey::config(lon,false);
          if template == "" {
              println!("{}", random.get_key());
          } else {
            for ky in random.from_template(template, lon) {
              println!("{}",ky);
            }
          };
        },
        Err(e) => println!("Error number conversion {}",e),
      };
    }
    return Ok(());
  }
//  if core::option::Option::is_some(&opts.tera) {
  if let Some(tera) = opts.tera {
    if let (Some(data_file),Some(tpl_file))= (tera.get(0),tera.get(1)) {
      let mut data_hash: HashMap<String, String> = HashMap::new();
      let mut tpl_context = tera::Context::new();
      // println!("data: {} template: {}",data_file,tpl_file);
      match hash_from_data(data_file, &mut tpl_context, &mut data_hash, true) {
        Ok(_) => {
         match data_templated(&"".to_string(),tpl_file,&mut tpl_context, &data_hash) {
           Ok(result) => println!("{}",result),
           Err(e) => println!("Error with {} data in {} template: {}", data_file, tpl_file, e),
         }
        },
        Err(e) => println!("Error with {} data in {} template: {}", data_file, tpl_file, e),
      }
      return Ok(());
    }
  }
  if let Some(text) = opts.hash {
    let res = hash_content(&text);
    println!("{}",res);
    return Ok(());
  }
  if let Some(_) = opts.uuid {
    println!("{}", uuid::Uuid::new_v4());
    return Ok(());
  }
  */
  // if let Some(_) = opts.http {
  //   let _ =http_main().await;
  //   return Ok(());
  // }
  /*
    for inpath in opts.inpath {
//        todo!("Implement application logic")
        println!("Target: {}",inpath.display());
    }

    */
  Ok(empty_res)
}

// Tests go below the code where they'll be out of the way when not the target of attention
#[cfg(test)]
mod tests {
    use super::CliOpts;

    // TODO: Unit test to verify that the doc comments on `CliOpts` or `BoilerplateOpts` aren't
    // overriding the intended about string.

    #[test]
    /// Test something
    fn test_something() {
        // TODO: Test something
    }
}
